from __future__ import annotations

from typing import Literal, Optional

from pydantic import BaseModel, NonNegativeInt, PositiveInt
from typing_extensions import NotRequired, TypedDict

from .common import AnnotationFieldName, EntityType


class PageParams(BaseModel):
    """
    Pagination params accepted by endpoints that return multiple
    resources
    """

    size: Optional[PositiveInt] = None
    """Number of results per page"""
    page: Optional[NonNegativeInt] = None
    """Which page to fetch (starting at 0)"""


# TODO: unclear whether the API accepts the same set of fields
#   for query/return. Defining them separately for now
# NOTE: fields seem to be badly documented. The best canonical source I've
#   been able to find seems to be going directly to the SOLR config for OLS
#   at https://github.com/EBISPOT/OLS/blob/dev/ols-solr/src/main/solr-5-config/ontology/conf/schema.xml
SearchReturnFields = Literal[
    "annotations",  # TODO: not sure if annotations does anything, annotations_trimmed does
    "annotations_trimmed",
    "description",
    "iri",
    "label",
    "obo_id",
    "ontology_name",
    "ontology_prefix",
    "short_form",
    "subset",
    "synonym",
    "type",
]
SearchQueryFields = Literal[
    "annotations",
    "description",
    "iri",
    "label",
    "logical_description",
    "obo_id",
    "short_form",
    "subset",
    "synonym",
]


class SearchParams(TypedDict):
    """
    Optional parameters passed to search() method (not including
    the q/query parameter).

    NOTE: use get_query_dict() to convert this to the format
    needed by the GET request
    """

    ontology: NotRequired[str | list[str]]
    """Ontologies to search, e.g. `["mondo", "upheno"]`"""
    type: NotRequired[EntityType]
    """Type of term to search for, e.g. "class", "property" """
    slim: NotRequired[list[str]]
    fieldList: NotRequired[list[SearchReturnFields | AnnotationFieldName]]
    """Which fields to return in the results"""
    queryFields: NotRequired[list[SearchQueryFields | AnnotationFieldName]]
    """Which fields to search over"""
    exact: NotRequired[bool]
    """Only return exact matches"""
    groupField: NotRequired[bool]
    """Group results by unique ID"""
    obsoletes: NotRequired[bool]
    """Include obsoleted terms in the results"""
    local: NotRequired[bool]
    """Only return terms in a defining ontology"""
    childrenOf: NotRequired[list[str]]
    """Restrict results to children of these terms"""
    allChildrenOf: NotRequired[list[str]]
    """
    Restrict results to children of these terms, plus other
    child-like relations e.g. "part of", "develops from"
    """
    rows: NotRequired[int]
    """
    Number of results per page
    """
    start: NotRequired[int]
    """
    Index of first result
    """


class SelectParams(TypedDict):
    """
    Optional parameters passed to select() method (not including
    the q/query parameter).

    NOTE: use get_query_dict() to convert this to the format
    needed by the GET request
    """

    ontology: NotRequired[str | list[str]]
    """Ontologies to search, e.g. `["mondo", "upheno"]`"""
    type: NotRequired[EntityType]
    """Type of term to search for, e.g. "class", "property" """
    slim: NotRequired[list[str]]
    fieldList: NotRequired[list[SearchReturnFields | AnnotationFieldName]]
    """Which fields to return in the results"""
    exact: NotRequired[bool]
    """Only return exact matches"""
    groupField: NotRequired[bool]
    """Group results by unique ID"""
    obsoletes: NotRequired[bool]
    """Include obsoleted terms in the results"""
    local: NotRequired[bool]
    """Only return terms in a defining ontology"""
    childrenOf: NotRequired[list[str]]
    """Restrict results to children of these terms"""
    allChildrenOf: NotRequired[list[str]]
    """
    Restrict results to children of these terms, plus other
    child-like relations e.g. "part of", "develops from"
    """
    rows: NotRequired[int]
    """
    Number of results per page
    """
    start: NotRequired[int]
    """
    Index of first result
    """


def get_query_dict(params: SearchParams | SelectParams) -> dict[str, str]:
    """
    Convert SearchParams or SelectParams to the format needed in requests,
    converting any list values to comma-separated string, as required by the search
    and select endpoints
    """
    query_dict = {}
    for field_name, value in params.items():
        if isinstance(value, list):
            value = ",".join(value)
        else:
            value = str(value)
        query_dict[field_name] = value
    return query_dict


RelativeTypes = Literal[
    "parents",
    "hierarchicalParents",
    "children",
    "hierarchicalChildren",
    "ancestors",
    "descendants",
    "hierarchicalDescendants",
    "hierarchicalAncestors",
]


class TermInDefiningOntologyParams(TypedDict):
    """
    Optional arguments for get_term_in_defining_ontology() method
    """

    iri: NotRequired[str]
    short_form: NotRequired[str]
    obo_id: NotRequired[str]
    id: NotRequired[str]


class GetTermsParams(TypedDict):
    """
    Optional arguments for get_terms() method
    """

    iri: NotRequired[str]
    short_form: NotRequired[str]
    obo_id: NotRequired[str]
    page: NotRequired[int]
    size: NotRequired[int]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pydantic import StringConstraints
from typing_extensions import Annotated

EntityType = Literal["class", "property", "individual", "ontology"]

# MyPy doesn't play nice with pydantic's constrained types,
#   see https://github.com/pydantic/pydantic/issues/3080
if TYPE_CHECKING:
    AnnotationFieldName = str
else:
    AnnotationFieldName = Annotated[str, StringConstraints(pattern=r"^\w+_annotation$")]



from __future__ import annotations

from typing import Any, Mapping, Optional, Type, TypeVar
from urllib.parse import quote_plus

import pydantic
import requests
from pydantic import validate_call

from . import schemas
from .instances import EBI_OLS4
from .schemas.requests import get_query_dict

S = TypeVar("S", bound=pydantic.BaseModel, covariant=True)
ParamsMapping = Mapping[str, Any]


class Ols4Client:
    """
    Client for communicating with an OLS instance.
    """

    base_url: str

    def __init__(self, base_url: str = EBI_OLS4):
        """
        :param base_url: Base API URL for the OLS instance, up to and including /api/
        """
        if not base_url.endswith("/"):
            base_url = base_url + "/"
        self.base_url = base_url
        self._session = requests.Session()
        self._session.headers.update({"accept": "application/json"})
        # TODO: do we need to set access-control-allow-origin header?

    def _create_url(self, path: str) -> str:
        # Remove leading /
        path = path.lstrip("/")
        return self.base_url + path

    @staticmethod
    def _quote_iri(iri: str) -> str:
        """
        Quote an IRI as a double URL-encoded URL string, so it can
        be used in a URL path
        :param iri: IRI, e.g. http://purl.obolibrary.org/obo/GO_0043226
        :return: Percent-encoded string, e.g. http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252FGO_0043226
        """
        return quote_plus(quote_plus(iri))

    def get(self, path: str, params: Optional[ParamsMapping] = None) -> dict:
        """
        Perform a GET request to the API. Unlike most of
        the client methods, this just returns the JSON data as a dict,
        without validation/parsing. Useful if you want to test
        an API endpoint, or use an endpoint that's not implemented yet.

        :param path: API path (excluding base url)
        :param params: Query parameters
        :return: JSON data, as a dict
        :raises HTTPError: if response is not OK
        """
        url = self._create_url(path)
        resp = self._session.get(url=url, params=params)
        resp.raise_for_status()
        json_data: dict = resp.json()
        return json_data

    def get_with_schema(
        self, schema: Type[S], path: str, params: Optional[ParamsMapping] = None
    ) -> S:
        """
        Get data from ``path`` and parse it with ``schema`` to return
        a pydantic object.

        :param schema: Pydantic class/model inheriting from BaseModel
        :param path: API path (excluding the base API url)
        :param params: Query parameters
        :return: Pydantic model instance created from ``schema``
        :raises pydantic.ValidationError: if response data fails
           to validate.
        """
        resp = self.get(path=path, params=params)
        obj = schema(**resp)
        return obj

    def get_api_info(self) -> schemas.responses.ApiInfo:
        """
        Get the list of endpoints supported by the API
        """
        resp = self.get_with_schema(schemas.responses.ApiInfo, path="/")
        return resp

    def get_ontologies(
        self, page: Optional[int] = None, size: Optional[int] = None
    ) -> schemas.responses.OntologyList:
        """
        Get the list of ontologies the OLS instance has.

        :param page: Page number of results (starting at 0)
        :param size: Number of results per page (API default is 20)
        """
        params = schemas.requests.PageParams(page=page, size=size).model_dump(
            exclude_none=True
        )
        ontology_list = self.get_with_schema(
            schemas.responses.OntologyList, "/ontologies", params=params
        )
        return ontology_list

    def get_ontology(self, ontology_id: str) -> schemas.responses.OntologyItem:
        """
        Get details for a single ontology

        :param ontology_id: Ontology ID/name, e.g. "mondo"
        """
        path = f"/ontologies/{ontology_id}/"
        ontology_item = self.get_with_schema(schemas.responses.OntologyItem, path)
        return ontology_item

    def get_term(self, ontology_id: str, iri: str) -> schemas.responses.Term:
        """
        Get a single term in a specific ontology

        :param ontology_id: Ontology ID/name, e.g. "mondo"
        :param iri: IRI for a single term
        :return: Term details
        """
        iri = self._quote_iri(iri)
        path = f"/ontologies/{ontology_id}/terms/{iri}"
        term = self.get_with_schema(schemas.responses.Term, path)
        return term

    def get_terms(
        self, ontology_id: str, params: Optional[schemas.requests.GetTermsParams] = None
    ) -> schemas.responses.MultipleTerms:
        """
        Get multiple terms in a specific ontology, possibly filtering by iri, short_form or obo_id.

        :param params: Optional params. Filter by iri, short_form or obo_id, or specify the
            number of results with ``page`` and ``size``
        """
        path = f"/ontologies/{ontology_id}/terms"
        return self.get_with_schema(
            schemas.responses.MultipleTerms, path, params=params
        )

    def find_terms(self, params: schemas.requests.GetTermsParams):
        """
        Search for terms across ontologies.

        Provide either `iri`, `short_form`, or `obo_id`

        :param params: Request params. You must provide one of `iri`, `short_form` or `obo_id`

        Example requests:

            curl -L 'http://www.ebi.ac.uk/ols4/api/terms/http%253A%252F%252Fwww.ebi.ac.uk%252Fefo%252FEFO_0000001' -i -H 'Accept: application/json'
            curl -L 'http://www.ebi.ac.uk/ols4/api/terms?iri=http://www.ebi.ac.uk/efo/EFO_0000001' -i -H 'Accept: application/json'
            curl -L 'http://www.ebi.ac.uk/ols4/api/terms?short_form=EFO_0000001' -i -H 'Accept: application/json'
            curl -L 'http://www.ebi.ac.uk/ols4/api/terms?obo_id=EFO:0000001' -i -H 'Accept: application/json'
            curl -L 'http://www.ebi.ac.uk/ols4/api/terms?id=EFO:0000001' -i -H 'Accept: application/json'
        """
        path = "/terms"
        return self.get_with_schema(
            schemas.responses.MultipleTerms, path, params=params
        )

    def get_term_in_defining_ontology(
        self,
        iri: Optional[str] = None,
        params: Optional[schemas.requests.TermInDefiningOntologyParams] = None,
    ) -> schemas.responses.TermInDefiningOntology:
        """
        Use the /terms/findByIdAndIsDefiningOntology/ to find a term in
        its defining ontology. This allows you to look up a term by IRI
        alone.

        :param iri: IRI for the term. You can either provide this, or use the ``params``
           argument (not both).
        :param params: GET parameters. the /findByIdAndIsDefiningOntology/ endpoint
           allows "iri", "short_form", "obo_id", or "id"
        :return: JSON data. Terms are at resp.embedded.terms
        :raises ValueError: if both/neither IRI and params arguments were given.
        """
        if iri and params:
            raise ValueError("Pass either iri or params arguments, not both")
        if iri:
            iri_encoded = self._quote_iri(iri)
            path = f"/terms/findByIdAndIsDefiningOntology/{iri_encoded}"
            return self.get_with_schema(
                schemas.responses.TermInDefiningOntology, path=path
            )
        if params:
            path = "/terms/findByIdAndIsDefiningOntology"
            return self.get_with_schema(
                schemas.responses.TermInDefiningOntology, path=path, params=params
            )
        raise ValueError("One of iri or params arguments is required")

    def _get_term_relatives(
        self, relatives: schemas.requests.RelativeTypes, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Common method for getting a term's parents, children, ancestors etc.
        """
        path = f"/ontologies/{ontology_id}/{relatives}"
        return self.get_with_schema(
            schemas.responses.MultipleTerms, path, params={"id": term_id}
        )

    def get_term_parents(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get parents for a term.

        :param ontology_id: Name of ontology, e.g. "go"
        :param term_id: Term ID (URI, short form or obo ID)
        :return: response object. the actual terms are in a list at
          ``response.embedded.terms``
        """
        return self._get_term_relatives(
            "parents", ontology_id=ontology_id, term_id=term_id
        )

    def get_term_hierarchical_parents(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get hierarchical parents for a term. See [get_term_parents()][ols_py.client.OlsClient.get_term_parents]
        """
        return self._get_term_relatives(
            "hierarchicalParents", ontology_id=ontology_id, term_id=term_id
        )

    def get_term_children(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get children for a term. See [get_term_parents()][ols_py.client.OlsClient.get_term_parents]
        """
        return self._get_term_relatives(
            "children", ontology_id=ontology_id, term_id=term_id
        )

    def get_term_ancestors(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get ancestors for a term. See [get_term_parents()][ols_py.client.OlsClient.get_term_parents]
        """
        return self._get_term_relatives(
            "ancestors", ontology_id=ontology_id, term_id=term_id
        )

    def get_term_descendants(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get descendants for a term. See [get_term_parents()][ols_py.client.OlsClient.get_term_parents]
        """
        return self._get_term_relatives(
            "descendants", ontology_id=ontology_id, term_id=term_id
        )

    def get_term_hierarchical_ancestors(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get hierarchical ancestors for a term. See [get_term_parents()][ols_py.client.OlsClient.get_term_parents]
        """
        return self._get_term_relatives(
            "hierarchicalAncestors", ontology_id=ontology_id, term_id=term_id
        )

    def get_term_hierarchical_descendants(
        self, ontology_id: str, term_id: str
    ) -> schemas.responses.MultipleTerms:
        """
        Get hierarchical descendants for a term. See [get_term_parents()][ols_py.client.OlsClient.get_term_parents]
        """
        return self._get_term_relatives(
            "hierarchicalAncestors", ontology_id=ontology_id, term_id=term_id
        )

    @staticmethod
    def _add_wildcards(query: str) -> str:
        with_wildcards = [f"{term}*" for term in query.split(" ")]
        return " ".join(with_wildcards)

    @validate_call
    def search(
        self,
        query: str,
        params: Optional[schemas.requests.SearchParams] = None,
        add_wildcards: bool = False,
    ) -> schemas.responses.SearchResponse:
        """
        Search for ``query`` using the /search API endpoint.

        :param query: term(s) to search for
        :param params: dictionary of search parameters
        :param add_wildcards: Add a wildcard * to each word in ``query`` -
           good for broad/flexible searches
        :return:
        """
        if add_wildcards:
            query = self._add_wildcards(query)
        if params is None:
            request_params = {"q": query}
        else:
            request_params = {"q": query, **get_query_dict(params)}
        resp = self.get_with_schema(
            schemas.responses.SearchResponse, "/search", params=request_params
        )
        return resp

    @validate_call
    def select(
        self,
        query: str,
        params: Optional[schemas.requests.SelectParams] = None,
        add_wildcards: bool = False,
    ) -> schemas.responses.SearchResponse:
        """
        Search for ``query`` using the /select API endpoint, which
        is supposed to be tuned to return good results for autocomplete.

        :param query: term(s) to search for
        :param params: dictionary of optional parameters
        :param add_wildcards: Add a wildcard * to each word in ``query`` -
           good for broad/flexible searches
        :return:
        """
        if add_wildcards:
            query = self._add_wildcards(query)
        if params is None:
            request_params = {"q": query}
        else:
            request_params = {"q": query, **get_query_dict(params)}
        resp = self.get_with_schema(
            schemas.responses.SearchResponse, "/search", params=request_params
        )
        return resp

    def get_property(self, ontology_id: str, iri: str) -> schemas.responses.Term:
        """
        Get a property from a specific ontology.

        Example request:

            curl -L 'http://www.ebi.ac.uk/ols4/api/ontologies/efo/properties/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252FBFO_0000050' -i -H 'Accept: application/json'
        """
        quoted_iri = self._quote_iri(iri)
        path = f"/ontologies/{ontology_id}/properties/{quoted_iri}"
        resp = self.get_with_schema(schemas.responses.Term, path=path)
        return resp

    def get_individual(self, ontology_id: str, iri: str) -> schemas.responses.Term:
        """
        Get an individual from a specific ontology.

        Example request:

            curl -L 'http://www.ebi.ac.uk/ols4/api/ontologies/iao/individuals/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252FIAO_0000002' -i -H 'Accept: application/json'
        """
        quoted_iri = self._quote_iri(iri)
        path = f"/ontologies/{ontology_id}/individuals/{quoted_iri}"
        resp = self.get_with_schema(schemas.responses.Term, path=path)
        return resp

    def get_related_term_by_property(
        self, ontology_id: str, term_iri: str, property_iri: str
    ):
        """
        Use the /ontologies/{ontology_id}/terms/{term_iri}/{property_iri} endpoint to find
        related terms.

        From the OLS4 docs:
            In cases where a term has a direct relation to another term (single existential to a
            named class in OBO), for example a "part of" relation, the related terms can be
            accessed directly with this API.

        Example request:

        http://www.ebi.ac.uk/ols4/api/ontologies/uberon/terms/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252FUBERON_0000016/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252FBFO_0000050
        """
        term_iri = self._quote_iri(term_iri)
        property_iri = self._quote_iri(property_iri)
        path = f"/ontologies/{ontology_id}/terms/{term_iri}/{property_iri}"
        return self.get_with_schema(schemas.responses.MultipleTerms, path=path)



from __future__ import annotations

from typing import Any, Optional

import pydantic
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, HttpUrl

from ols_py.schemas.common import EntityType


class PageInfo(BaseModel):
    """
    Page information returned in paginated responses
    """

    size: int
    totalElements: int
    totalPages: int
    number: int


class Link(BaseModel):
    """
    Link item returned in responses
    """

    href: HttpUrl


# TODO: not sure which of these are optional
class OboXref(BaseModel):
    database: Optional[str] = None
    id: Optional[str] = None
    description: Optional[str] = None
    url: Optional[str] = None


class OboSynonym(BaseModel):
    name: str
    # TODO: can this be a fixed set, e.g. hasExactSynonym, hasRelatedSynonym?
    scope: str
    type: Optional[str] = None
    xrefs: list[OboXref] = Field(default_factory=list)


class Term(BaseModel):
    """
    Response returned by term endpoints
    """

    # Haven't typed all attributes yet, so allow additional fields
    #   for now
    model_config = ConfigDict(extra="allow")
    iri: pydantic.AnyUrl
    label: str
    description: list[str]
    # Not specifying annotations for now, not sure if these
    #   are fixed
    annotation: dict[str, list[str]]
    synonyms: Optional[list[str]] = None
    obo_xref: Optional[list[OboXref]] = None
    obo_synonym: Optional[list[OboSynonym]] = None
    ontology_name: str
    ontology_prefix: str
    ontology_iri: pydantic.AnyUrl
    is_obsolete: bool
    term_replaced_by: Optional[Any] = None
    has_children: bool
    is_root: bool
    short_form: str
    # Higher level terms may not have an obo ID, e.g.
    # terms will be descendants of http://www.w3.org/2002/07/owl#Thing
    obo_id: Optional[str] = None
    in_subset: Optional[Any] = None
    links: dict[str, Link] = Field(..., alias="_links")


class ApiInfoLinks(BaseModel):
    """
    Set of links returned in the root endpoint/
    API ifno
    """

    ontologies: Link
    individuals: Link
    terms: Link
    properties: Link
    profile: Link


class ApiInfo(BaseModel):
    """
    Response returned by the root API endpoint,
    links to other endpoints/resources
    """

    links: ApiInfoLinks = Field(..., alias="_links")


class OntologyItemLinks(BaseModel):
    self: Link
    terms: Link
    properties: Link
    individuals: Link


class OntologyItem(BaseModel):
    ontologyId: str
    status: str
    numberOfProperties: int
    numberOfTerms: int
    languages: Optional[list[str]] = None
    links: OntologyItemLinks = Field(..., alias="_links")

    # TODO: not all fields have been documented so far, allow
    #   them through with extra="allow" for now
    model_config = ConfigDict(extra="allow")


class OntologyListEmbedded(BaseModel):
    ontologies: list[OntologyItem]


class OntologyList(BaseModel):
    page: PageInfo
    embedded: OntologyListEmbedded = Field(None, alias="_embedded")


class EmbeddedTerms(BaseModel):
    """
    "_embedded" field used in responses returning terms
    """

    terms: list[Term]


class TermRelativesLinks(BaseModel):
    self: Link


class MultipleTerms(BaseModel):
    """
    Response returned for endpoints which return multiple term results,
    e.g. parents, ancestors, descendants etc.
    The actual terms are at ``response.embedded.terms``
    """

    embedded: EmbeddedTerms = Field(None, alias="_embedded")
    links: TermRelativesLinks = Field(None, alias="_links")
    page: PageInfo


class TermInDefiningOntologyLinks(BaseModel):
    self: Link


class TermInDefiningOntology(BaseModel):
    """
    Response returned for /terms/findByIdAndIsDefiningOntology/
    endpoint
    """

    embedded: EmbeddedTerms | None = Field(None, alias="_embedded")
    links: TermInDefiningOntologyLinks = Field(..., alias="_links")
    page: PageInfo


class SearchResultItem(BaseModel, extra="allow"):
    id: Optional[str] = None
    annotations: Optional[list[str]] = None
    annotations_trimmed: Optional[list[str]] = None
    description: Optional[list[str]] = None
    iri: Optional[str] = None
    label: Optional[str] = None
    obo_id: Optional[str] = None
    ontology_name: Optional[str] = None
    ontology_prefix: Optional[str] = None
    subset: Optional[list[str]] = None
    short_form: Optional[str] = None
    # This field is actually named 'synonym' in OLS search
    #   responses, but we remap it to synonyms
    synonyms: Optional[list[str]] = Field(
        default=None, validation_alias=AliasChoices("synonyms", "synonym")
    )
    type: Optional[EntityType] = None


class SearchResponseResponse(BaseModel):
    numFound: int
    start: int
    docs: list[SearchResultItem]


class SearchResponse(BaseModel):
    responseHeader: dict
    response: SearchResponseResponse


class OlsErrorSchema(BaseModel):
    """
    Error data returned the OLS API for a bad request/error
    """

    error: str
    message: str
    path: str
    status: int
    timestamp: int | float
